! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!! A section of a 3D domain.
!!
!! The section is made up of a very limited number of clusters each consisting
!! of a 3D rectangle (cuboid).
!!
!! Rectangles can be added to a section, the algorithm tries to reduce the total
!! volume of all clusters.
!!
!! @author Matthias Lieber
module section_mod
  
  use kinds_mod
  implicit none
  
  private
  
  integer, parameter :: SECTION_RECT_MAX = 4
  
  !! A 3D section consisting of a limited number of 3D cuboids
  type section
    integer(i_k) :: rect(3,2,SECTION_RECT_MAX)  !! a limited number of 3D cuboids
    integer(i_k) :: nrect = 0                   !! number of used cuboids
    integer(i_k) :: maxext(3)
  end type section

  public :: section, section_reset, section_add, section_intersect, section_volume

contains

  
  !! Reset $sect (set volume to null).
  !!
  !! Parameter $maxrect will be used in $section_add to estimate if it
  !! is useful to add a new cuboid or not.
  !!
  subroutine section_reset(sect, maxrect)
    type(section), intent(inout) :: sect         !! the section
    integer(i_k),  intent(in) :: maxrect(3,2)    !! max allowed size 
    
    sect%nrect = 0
    ! spans a negative section = null
    sect%rect(1:3,1,1) = maxrect(1:3,2) + 1
    sect%rect(1:3,2,1) = maxrect(1:3,1) - 1
    ! save maxext
    sect%maxext = maxrect(1:3,2) - maxrect(1:3,1) + 1
    
  end subroutine section_reset


  !! Add $rect to section $sect (union).
  !!
  !! If necessary it will create a new cuboid in $sect to keep overall
  !! volume as small as possible.
  !!
  subroutine section_add(sect, rect)
    type(section), intent(inout) :: sect         !! the section
    integer(i_k),  intent(in) :: rect(3,2)       !! 3D cuboid to add
    
    integer(i_k) :: tmp(3,2), irect
    integer(i_k) :: ext1(3), ext2(3)
    integer(i_k) :: hit
    integer(i_k) :: v1(SECTION_RECT_MAX), v2(SECTION_RECT_MAX)

   
    if(sect%nrect <= 0) then
    
      !! special case: empty $sect
      
      sect%nrect = 1
      sect%rect(1:3,1:2,1) = rect(1:3,1:2)
      
    elseif(sect%nrect < SECTION_RECT_MAX) then
      
      !! we could still create a new rect
      
      hit = 1
      do irect = 1, sect%nrect
        hit = irect
        ! unify sect%rect with rect
        tmp(1:3,1) = MIN( sect%rect(1:3,1,irect), rect(1:3,1) )
        tmp(1:3,2) = MAX( sect%rect(1:3,2,irect), rect(1:3,2) )
        ! calculate extents (before/after union)
        ext1(1:3) = sect%rect(1:3,2,irect) - sect%rect(1:3,1,irect) + 1
        ext2(1:3) = tmp(1:3,2) - tmp(1:3,1) + 1
        ! relative growth in any dimension > 2 ?
        ! and size of union in this dimension > max size / 4 ?
        if( ANY(ext2 > ext1 * 2 .and. ext2 * 4 > sect%maxext) ) then
            ! this is no good rect for the union, skip this and try next one
            hit = 0
        end if
        ! ok, we found a good rect, let's get out of here
        if(hit /= 0) exit      
      end do


      if(hit /= 0) then
        ! found good rect for union
        sect%rect(1:3,1:2,hit) = tmp(1:3,1:2)
      else
        ! found no good rect
        ! but we can still create a new one
        sect%nrect = sect%nrect + 1
        sect%rect(1:3,1:2,sect%nrect) = rect(1:3,1:2)
      end if

    else
    
      !! we cannot create a new rect
      !! so just add this rect in such a way, that the sum volume is smallest

      do irect = 1, sect%nrect
        ! unify sect%rect with rect
        tmp(1:3,1) = MIN( sect%rect(1:3,1,irect), rect(1:3,1) )
        tmp(1:3,2) = MAX( sect%rect(1:3,2,irect), rect(1:3,2) )
        ! calculate extents (before/after union)
        ext1(1:3) = sect%rect(1:3,2,irect) - sect%rect(1:3,1,irect) + 1
        ext2(1:3) = tmp(1:3,2) - tmp(1:3,1) + 1
        ! volumes
        v1(irect) = PRODUCT(ext1)
        v2(irect) = PRODUCT(ext2)        
      end do

      ! calculate in v2 the total volume of all cuboids after adding to sect%rect(irect)
      do irect = 1, sect%nrect
        v2(irect) = v2(irect) + SUM(v1) - v1(irect)
      end do
      
      ! ok, this must be it
      hit = MINLOC(v2,1)
      sect%rect(1:3,1,hit) = MIN( sect%rect(1:3,1,hit), rect(1:3,1) )
      sect%rect(1:3,2,hit) = MAX( sect%rect(1:3,2,hit), rect(1:3,2) )      
      
    end if    
        
    
  end subroutine section_add


  !! Return .true. if intersection of $sect with cuboid $rect is not empty.
  !!
  !! Optionally, an offset used for the upper bound of $sect can be given.
  !!
  logical function section_intersect(sect, rect, opt_offset)
    type(section), intent(inout) :: sect         !! the section
    integer(i_k),  intent(in) :: rect(3,2)       !! 3D cuboid to intersect
    integer(i_k),  intent(in) :: opt_offset(3)   !! optional offset for upper bound of $sect
    optional :: opt_offset
    
    integer(i_k) :: intersec(3,2), irect

    section_intersect = .false.
    
    if(present(opt_offset)) then
    
      do irect = 1, sect%nrect
        ! intersect sect%rect with rect
        intersec(1:3,1) = MAX( sect%rect(1:3,1,irect)                , rect(1:3,1) )
        intersec(1:3,2) = MIN( sect%rect(1:3,2,irect)+opt_offset(1:3), rect(1:3,2) )
        
        !write(*,'(2(I3),A,3(I5),A,3(I5))') irect,sect%nrect,': ',sect%rect(1:3,1,irect), ' - ',sect%rect(1:3,2,irect)
        
        if( ALL( intersec(1:3,1) <= intersec(1:3,2) ) ) then
          ! intersection is not empty
          section_intersect = .true.
          return
        end if
      end do
    
    else

      do irect = 1, sect%nrect
        ! intersect sect%rect with rect
        intersec(1:3,1) = MAX( sect%rect(1:3,1,irect), rect(1:3,1) )
        intersec(1:3,2) = MIN( sect%rect(1:3,2,irect), rect(1:3,2) )
        if( ALL( intersec(1:3,1) <= intersec(1:3,2) ) ) then
          ! intersection is not empty
          section_intersect = .true.
          return
        end if
      end do
      
    end if
    
  end function section_intersect


  !! Calculate total volume of the section.
  !!
  subroutine section_volume(sect, volume)
    type(section), intent(inout) :: sect   !! the section
    integer(i_k),  intent(out) :: volume   !! volume
    integer(i_k) :: irect
      
    volume = 0
    do irect = 1, sect%nrect
      volume = volume + PRODUCT( sect%rect(1:3,2,irect) - sect%rect(1:3,1,irect) + 1 )
    end do
  
  end subroutine section_volume


end module section_mod
